// Fifo Handling

struct mj_keybuffer {
  unsigned char c;  // the Character
  unsigned char d;  // the Delay in Bits 0-3, Bit 7 = "Special Char" which will not be auto-released
};

byte mj_actionstore[MJ_FIFO_A_SIZE];
int mj_fifo_a_index_nextout = 0;
int mj_fifo_a_index_nextin = 0;

mj_keybuffer mj_fifostore[MJ_FIFOSIZE];
int mj_fifoindex_nextout = 0;
int mj_fifoindex_nextin = 0;

// Get one fifo entry
byte mj_fifo_a_getelement() {
  int index;
  if (! (mj_fifo_a_index_nextout == mj_fifo_a_index_nextin)) {
    index = mj_fifo_a_index_nextout;
    if (mj_fifo_a_index_nextout == (MJ_FIFO_A_SIZE - 1)) {
      mj_fifo_a_index_nextout = 0;
    } else {
      mj_fifo_a_index_nextout++;
    }
    return mj_actionstore[index];
  } else {
    return (0);
  }
}

// Get one fifo entry but don't delete from fifo
byte mj_fifo_a_getelement_n() {
  if (! (mj_fifoindex_nextout == mj_fifo_a_index_nextin)) {
    return mj_actionstore[mj_fifo_a_index_nextout];
  } else {
    return (0);
  }
}

// tell how many entries are left in fifo
int mj_fifo_a_stored() {
  if (mj_fifo_a_index_nextin >= mj_fifo_a_index_nextout) {
    return (mj_fifo_a_index_nextin - mj_fifo_a_index_nextout);
  } else {
    return ((MJ_FIFO_A_SIZE - mj_fifo_a_index_nextout) + mj_fifo_a_index_nextin);
  }
}

// put an element into fifo
bool mj_fifo_a_putelement(byte b) {
  int index;
  index = mj_fifo_a_index_nextin;
  if (index == (MJ_FIFO_A_SIZE - 1)) {
    index = 0;
  } else {
    index++;
  }
  if (index == mj_fifo_a_index_nextout) {
    return (false);
  } else {
    mj_actionstore[mj_fifo_a_index_nextin] = b;
    mj_fifo_a_index_nextin = index;
  }
  return (true);
}

// Get one fifo entry
mj_keybuffer mj_fifo_getelement() {
  int index;
  if (! (mj_fifoindex_nextout == mj_fifoindex_nextin)) {
    index = mj_fifoindex_nextout;
    if (mj_fifoindex_nextout == (MJ_FIFOSIZE - 1)) {
      mj_fifoindex_nextout = 0;
    } else {
      mj_fifoindex_nextout++;
    }
    return mj_fifostore[index];;
  } else {
    return ((mj_keybuffer) {
      0, 0
    });
  }
}

// tell how many entries are left in fifo
int mj_fifo_stored() {
  if (mj_fifoindex_nextin >= mj_fifoindex_nextout) {
    return (mj_fifoindex_nextin - mj_fifoindex_nextout);
  } else {
    return ((MJ_FIFOSIZE - mj_fifoindex_nextout) + mj_fifoindex_nextin);
  }
}

// put an element into fifo
bool mj_fifo_putelement(mj_keybuffer b) {
  int index;
  index = mj_fifoindex_nextin;
  if (index == (MJ_FIFOSIZE - 1)) {
    index = 0;
  } else {
    index++;
  }
  if (index == mj_fifoindex_nextout) {
    return (false);
  } else {
    mj_fifostore[mj_fifoindex_nextin] = b;
    mj_fifoindex_nextin = index;
  }
  return (true);
}

// Now more MJ Specific Code - above is mostly generic.
#ifdef HID_PROJECT_VERSION
const unsigned char MJ_KEYPAD[] = { KEYPAD_0, KEYPAD_1, KEYPAD_2, KEYPAD_3, KEYPAD_4, KEYPAD_5, KEYPAD_6, KEYPAD_7, KEYPAD_8, KEYPAD_9 };
#else
const unsigned char MJ_KEYPAD[] = { 234, 225, 226, 227, 228, 229, 230, 231, 232, 233 };
#endif
const unsigned char MJ_TRANSLATE[] = {'!', '"', '#', '$', '%', '&', 39, '(', ')', '*',
                                      '+', ',', '-', '.', '/', ':', ';', '<', '=', '>',
                                      '?', '@', 'A', 'Q', 'Y', 'Z', '[', ']', '\\', '^',
                                      '_', '`', 'a', 'q', 'y', 'z', '{', '|', '}', '~'
                                     };

void putfifo(unsigned char c, unsigned char d) {
  unsigned char e;
  e = d & 0x0f;
  mj_fifo_putelement((mj_keybuffer) {
    c, e
  });
}

void putfifo_special(unsigned char c, unsigned char d) {
  unsigned char e;
  e = d | 0x80;
  mj_fifo_putelement((mj_keybuffer) {
    c, e
  });
}

#ifdef HID_PROJECT_VERSION
void putfifo_system(unsigned char c, unsigned char d) {
  unsigned char e;
  e = d | 0x40;
  mj_fifo_putelement((mj_keybuffer) {
    c, e
  });
}

void putfifo_keycode(unsigned char c, unsigned char d) {
  unsigned char e;
  e = d | 0x20;
  mj_fifo_putelement((mj_keybuffer) {
    c, e
  });
}
#endif

// Adds Character to HID Fifo and translate "problematic" Chars to their Alt-<ascii> ones which should work with nearly any key-layout on windows or DOS
void putfifo_translate(unsigned char c, int d) {
  unsigned int i;
  int found = 0;
  for (i = 0; i < sizeof(MJ_TRANSLATE); i++ ) {
    if (c == MJ_TRANSLATE[i]) {
      found = 1;
      break;
    }
  }
  if (found == 1) {
    if (mj_osmode == 0) {
#ifdef HID_PROJECT_VERSION
      putfifo_keycode(KEY_LEFT_ALT, 0x80);
#else
      putfifo_special(KEY_LEFT_ALT, 1);
#endif
      if (c >= 100) {
#ifdef HID_PROJECT_VERSION
        putfifo_keycode(MJ_KEYPAD[c / 100], 1);
#else
        putfifo(MJ_KEYPAD[c / 100], 1);
#endif
      }
#ifdef HID_PROJECT_VERSION
      putfifo_keycode(MJ_KEYPAD[(c / 10) % 10], 1);
      putfifo_keycode(MJ_KEYPAD[c % 10], 1);
#else
      putfifo(MJ_KEYPAD[(c / 10) % 10], 1);
      putfifo(MJ_KEYPAD[c % 10], 1);
#endif
      putfifo_special(KEY_RELEASE, d);
    } else {
#ifdef HID_PROJECT_VERSION
      putfifo_keycode(KEY_LEFT_CTRL, 0x80);
      putfifo_keycode(KEY_LEFT_SHIFT, 0x80);
      putfifo('u', 1);
#else
      putfifo_special(KEY_LEFT_CTRL, 0);
      putfifo_special(KEY_LEFT_SHIFT, 0);
      putfifo_special('u', 1);
#endif
      putfifo_special(KEY_RELEASE, 1);
      i = (c / 16);
      if (i < 10) {
        i = i + '0';
      } else {
        i = (i + 'a') - 10;
      }
      putfifo(i, 1);
      i = c & 0x0f;
      if (i < 10) {
        i = i + '0';
      } else {
        i = (i + 'a') - 10;
      }
      putfifo(i, 1);
#ifdef HID_PROJECT_VERSION
      putfifo_keycode(KEY_RETURN, 1);
#else
      putfifo(KEY_RETURN, 1);
#endif
    }
  } else {
    putfifo(c, d);
  }
}

// Put a string into fifo with interpreting it
// type = 1 for string in c and 0 for eeprom from offset MJ_EEPROM_CONTENT
void putfifo_string(char *c, int type) {
  unsigned int len = 0, i = 0;
  int d;
  unsigned char ch;
  d = mj_keydelay_default;
  // Determine length of string given - method differs for different string sources
  if (type > 0 && type <= 14) { // String in EEPROM
    len = mj_usedbytes[type - 1];
  } else if (type == 0) { // Normal String
    len = strlen((const char*) c);
  } else if (type == 255) { // String in PROGMEM
    for (i = 0; pgm_read_byte(c + i) > 0 && i < MJ_EEPROM_CONTENT_14_SIZE; i++) {
      ;
    }
    len = i;
  }
  if (len >= MJ_FIFOSIZE) {
    len = MJ_FIFOSIZE - 1;
  }
  for (i = 0; i < len; i++) {
    if ( mj_getvalue(i, type, c) == '\\' && (i + 1) < len ) {
      switch (mj_getvalue(i + 1, type, c)) {
        case 'e': // ESC
#ifdef HID_PROJECT_VERSION
          putfifo_keycode(KEY_ESC, d);
#else
          putfifo(KEY_ESC, d);
#endif
          break;
        case 'n': // RETURN
        case 'r':
#ifdef HID_PROJECT_VERSION
        case '\n':
          putfifo_keycode(KEY_RETURN, d);
#else
          putfifo(KEY_RETURN, d);
#endif
          break;
        case 't': // TAB
#ifdef HID_PROJECT_VERSION
          putfifo_keycode(KEY_TAB, d);
#else
          putfifo(KEY_TAB, d);
#endif
          break;
        case '\\': // '\'
          putfifo_translate('\\', d);
          break;
        case 0x27: // '
          putfifo_translate(0x27, d);
          break;
        case '"':
          putfifo_translate('"', d);
          break;
        case '?':
          putfifo_translate('?', d);
          break;
        case '0':
          putfifo_translate(0, d);
          break;
        case 'd': // Set new delay for following keys
          if ((i + 2) < len) {
            ch = mj_getvalue(i + 2, type, c);
            if ( ch >= '0' && ch <= '9' ) {
              d = (int) ch - '0';
            }
          }
          i++;
          break;
        case 'x': // Hex Code - invalid Codes will be silently skipped (Not 0-9/a-f/A-F)
          if ((i + 3) < len) {
            putfifo_translate(mj_convert_hex_char( mj_getvalue(i + 2, type, c) , mj_getvalue(i + 3, type, c)), d);
            i += 2;
          }
          break;
        case 'X': // Hex Code - like above but with no translation
          if ((i + 3) < len) {
#ifdef HID_PROJECT_VERSION
            putfifo_keycode(mj_convert_hex_char( mj_getvalue(i + 2, type, c) , mj_getvalue(i + 3, type, c)), d);
#else
            putfifo(mj_convert_hex_char( mj_getvalue(i + 2, type, c) , mj_getvalue(i + 3, type, c)), d);
#endif
            i += 2;
          }
          break;
        case 's': // Hex Code - like above but for "Special Keys" (Only Press, no Auto-Release - \sff = Release all Keys
          if ((i + 3) < len) {
#ifdef HID_PROJECT_VERSION
            putfifo_special(mj_convert_hex_char( mj_getvalue(i + 2, type, c) , mj_getvalue(i + 3, type, c)), d | 0x20 );
#else
            putfifo_special(mj_convert_hex_char( mj_getvalue(i + 2, type, c) , mj_getvalue(i + 3, type, c)), d);
#endif
            i += 2;
          }
          break;
#ifdef HID_PROJECT_VERSION
        case 'S': // 1-Digit-Hex Code - like above but for "System Keys"
          if ((i + 2) < len) {
            putfifo_system(mj_convert_hex_nibble( mj_getvalue(i + 2, type, c)), d);
            i += 1;
          }
          break;
#endif
        default:
          putfifo_translate('\\', d);
          putfifo_translate(mj_getvalue(i + 1, type, c), d);
          break;
      }
      i++;
    } else {
      putfifo_translate(mj_getvalue(i, type, c), d);
    }
  }
}
